/**
 * Copyright (c) 2011, REMPL.com
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1) Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 * 2) Redistributions in binary form must reproduce the above
 * copyright notice, this list of conditions and the following
 * disclaimer in the documentation and/or other materials provided
 * with the distribution.
 * 3) Neither the name of the REMPL.com nor the names of its
 * contributors may be used to endorse or promote products derived
 * from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package com.rempl.readers.its;

// REMPL API from com.rempl:rempl-api classes
import com.rempl.api.Model;
import com.rempl.api.Reader;
import com.rempl.api.Rempler;
import com.rempl.api.RuntimeProblem;

// UML API
import com.rempl.api.uml.Diagram;
import com.rempl.api.uml.Instance;

// Bridge API from com.rempl.readers.its:bridge-its-api
import com.rempl.readers.its.bridge.Bridge;
import com.rempl.readers.its.bridge.BridgeException;
import com.rempl.readers.its.bridge.BridgeFactory;
import com.rempl.readers.its.bridge.Change;
import com.rempl.readers.its.bridge.ChangeAssignee;
import com.rempl.readers.its.bridge.ChangeMessage;
import com.rempl.readers.its.bridge.ChangeStatus;
import com.rempl.readers.its.bridge.ChangeTitle;
import com.rempl.readers.its.bridge.Issue;
import com.rempl.readers.its.bridge.Person;

// date processing
import java.text.SimpleDateFormat;

// collection management and dates
import java.util.Date;
import java.util.Locale;
import java.util.Map;

/**
 * Reader of ITS tickets/issues.
 *
 * @author Yegor Bugayenko (yegor@rempl.com)
 * @version $Id: ITSReader.java 1389 2011-05-23 23:07:59Z guard $
 * @todo #2 I don't think that we're using Java Generics effectively
 *       here. Actually generics are NOT used at all, but we can use them
 *       since that information is provided by ITS-BRIDGE.
 */
public final class ITSReader implements Reader {

    /**
     * Value attribute of a change.
     */
    private static final String ATTR_VALUE = "value";

    /**
     * The {@link Rempler} we're working for.
     * @see #setRempler(Rempler)
     */
    private Rempler rempler;

    /**
     * Public ctor.
     */
    public ITSReader() {
        this.rempler = null;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void setRempler(final Rempler rpr) {
        this.rempler = rpr;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public boolean read(final Model model) throws java.io.IOException {
        final Bridge bridge = BridgeFactory.INSTANCE.makeBridge(
            this.rempler.configuration().get("/*/its/bridge/text()")
        );
        try {
            bridge.configure(
                (Map<String, String>) this.rempler.configuration().get(
                    "/*/its/params/*",
                    Map.class
                )
            );
        } catch (BridgeException ex) {
            throw RuntimeProblem.make(ex);
        }
        Diagram diagram;
        try {
            diagram = this.diagram(bridge.retrieve());
        } catch (BridgeException ex) {
            throw RuntimeProblem.make(ex);
        }
        model.transform(diagram.uml());
        return false;
    }

    /**
     * Retrieve information from ITS issues and inject them
     * into the diagram.
     * @param issues The list of issues from ITS
     * @return The diagram just created
     * @throws BridgeException If something goes wrong
     */
    private Diagram diagram(final Map<String, Issue> issues)
        throws BridgeException {
        final Diagram diagram = new Diagram();
        for (Map.Entry<String, Issue> entry : issues.entrySet()) {
            final Instance issue = diagram.add("Issue");
            issue.add("name", entry.getKey());
            for (Change change : entry.getValue().changes()) {
                final Instance chg = diagram.add("IssueChange");
                issue.add("change", chg);
                chg.add("relatedIssue", issue);
                chg.add("time", this.date(change.time()));
                chg.add("author", this.person(diagram, change.author()));
                chg.add("component", this.component(change));
                this.value(diagram, chg, change);
            }
        }
        return diagram;
    }

    /**
     * Make a Person UML element.
     * @param diagram The diagram to use
     * @param person The person to analyze
     * @return The instance
     */
    private Instance person(final Diagram diagram, final Person person) {
        final Instance inst = diagram.add("Person");
        // inst.add("name", person.name());
        inst.add("email", person.email());
        return inst;
    }

    /**
     * Convert date to string.
     * @param time The time
     * @return The text
     */
    private String date(final Date time) {
        return new SimpleDateFormat(
            "yyyy-MM-dd'T'HH:mm:ssz",
            Locale.ENGLISH
        ).format(time);
    }

    /**
     * Make a title of the component.
     * @param change The change to analyze
     * @return The type
     */
    private String component(final Change change) {
        String type;
        if (change instanceof ChangeAssignee) {
            type = "assignee";
        } else if (change instanceof ChangeMessage) {
            type = "message";
        } else if (change instanceof ChangeStatus) {
            type = "status";
        } else if (change instanceof ChangeTitle) {
            type = "title";
        } else {
            throw RuntimeProblem.make(
                "Unknown type of change: %s",
                change.getClass().getName()
            );
        }
        return type;
    }

    /**
     * Inject value into change.
     * @param diagram The diagram to use
     * @param chg The UML instance to use
     * @param change The change to analyze
     */
    private void value(final Diagram diagram, final Instance chg,
        final Change change) {
        if (change instanceof ChangeAssignee) {
            chg.add(
                this.ATTR_VALUE,
                this.person(diagram, ((ChangeAssignee) change).assignee())
            );
        } else if (change instanceof ChangeMessage) {
            chg.add(this.ATTR_VALUE, ((ChangeMessage) change).message());
        } else if (change instanceof ChangeStatus) {
            final ChangeStatus.Status status =
                ((ChangeStatus) change).status();
            chg.add(this.ATTR_VALUE, status.toString());
        } else if (change instanceof ChangeTitle) {
            chg.add(this.ATTR_VALUE, ((ChangeTitle) change).title());
        } else {
            throw RuntimeProblem.make(
                "Unparseable type of change: %s",
                change.getClass().getName()
            );
        }
    }

}
